#include <hipc/msgd.h>

/**
 * Returns body size of a HIPC message data.
 *
 * @param md_p pointer to a HIPC message data
 * @returns body size of the HIPC message data
 */
unsigned char hipcMsgdBdySize(hipc_msgd const *const md_p)
{
    return hipcMsghdrBdysize(md_p->header);
}

/**
 * Returns type of a HIPC message data.
 *
 * @param md_p pointer to a HIPC message data
 * @returns type of the HIPC message data
 */
unsigned char hipcMsgdType(hipc_msgd const *const md_p)
{
    return md_p->header[HIPC_MSGHDR_BYTE_TYPE];
}

/**
 * Returns structure number of a HIPC message data.
 *
 * @param md_p pointer to a HIPC message data
 * @returns structure number of the HIPC message data
 */
unsigned char hipcMsgdStruno(hipc_msgd const *const md_p)
{
    return md_p->header[HIPC_MSGHDR_BYTE_STRUNO];
}

/**
 * Returns offset of a HIPC message data.
 *
 * @param md_p pointer to a HIPC message data
 * @returns offset of the HIPC message data
 */
unsigned char hipcMsgdOffset(hipc_msgd const *const md_p)
{
    return md_p->header[HIPC_MSGHDR_BYTE_OFFSET];
}

/**
 * Returns range size of a HIPC message data.
 *
 * @param md_p pointer to a HIPC message data
 * @returns range size of the HIPC message data
 */
unsigned char hipcMsgdRngsize(hipc_msgd const *const md_p)
{
    return md_p->header[HIPC_MSGHDR_BYTE_BDYSZ];
}

/**
 * Initializes a HIPC message data as a unstructured type message.
 *
 * @param md_p pointer to a HIPC message data
 * @param type message type
 * @param hipc_errno HIPC error number
 * @param errdtlstr error detail string
 * @param errdtlstr_size size of the error detail string
 * @returns the same value as \c hipc_errno.
 */
enum HIPC_errno hipcGenmdUnstru(hipc_msgd * const md_p,
                                const unsigned char type,
                                const enum HIPC_errno hipc_errno,
                                char *const errdtlstr,
                                size_t errdtlstr_size)
{
    md_p->header[HIPC_MSGHDR_BYTE_TYPE] = type;
    md_p->header[HIPC_MSGHDR_BYTE_OFFSET] = hipc_errno;
    md_p->header[HIPC_MSGHDR_BYTE_STRUNO] = HIPC_UNSTRU;
    md_p->body_p = errdtlstr;

    if (NULL == errdtlstr) {
        md_p->header[HIPC_MSGHDR_BYTE_BDYSZ] = 0;
    } else {
        if (255 < errdtlstr_size) {
            md_p->header[HIPC_MSGHDR_BYTE_BDYSZ] = 255;
        } else {
            md_p->header[HIPC_MSGHDR_BYTE_BDYSZ] = errdtlstr_size;
        }
    }

    return hipc_errno;
}

/**
 * Initializes \c mds_p as a response to \c mdr_p whose type must be GET.
 * \c stru_p should be the beginning of a structure specified by the GET message.
 *
 * @param mds_p pointer to a HIPC message data to be sent
 * @param mdr_p pointer to a HIPC message data received
 * @param stru_p pointer to a structure object
 * @param stru_size size of the structure object
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno hipcGenmdSuccessToGet(hipc_msgd * const mds_p,
                                      hipc_msgd const *const mdr_p,
                                      void *const stru_p,
                                      const size_t stru_size)
{
    if (stru_size < (size_t) mdr_p->header[HIPC_MSGHDR_BYTE_OFFSET] +
        mdr_p->header[HIPC_MSGHDR_BYTE_BDYSZ]) {
        return HIPC_ERR_S_INV_RW_RANGE;
    }

    mds_p->header[HIPC_MSGHDR_BYTE_TYPE] = HIPC_MESSAGE_TYPE_SUCCESS;
    mds_p->header[HIPC_MSGHDR_BYTE_OFFSET] =
        mdr_p->header[HIPC_MSGHDR_BYTE_OFFSET];
    mds_p->header[HIPC_MSGHDR_BYTE_STRUNO] =
        mdr_p->header[HIPC_MSGHDR_BYTE_STRUNO];
    mds_p->header[HIPC_MSGHDR_BYTE_BDYSZ] =
        mdr_p->header[HIPC_MSGHDR_BYTE_BDYSZ];
    mds_p->body_p =
        ((unsigned char *const) stru_p) +
        mdr_p->header[HIPC_MSGHDR_BYTE_OFFSET];

    return HIPC_SUCCESS;
}

/**
 * Sets a body to a HIPC message data that contains a header of a PUT message.
 * The body will be a part of \c stru_p. \c stru_p should be the beginning of
 * a structure which is specified by the header of \c md_p
 * so that the following statement reads the body of the PUT message
 * into correct place.
 *
 * hipcUsocRead(so_p, hipcMsgdBdyPtr(md_p), hipcMsgdBdySize(md_p));
 *
 * @param mdr_p pointer to a HIPC message data
 * @param stru_p pointer to a structure object
 * @param stru_size size of the structure object
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno hipcMsgdSetBdyForPut(hipc_msgd * const mdr_p,
                                     void *const stru_p,
                                     const size_t stru_size)
{
    if (stru_size < (size_t) mdr_p->header[HIPC_MSGHDR_BYTE_OFFSET] +
        mdr_p->header[HIPC_MSGHDR_BYTE_BDYSZ]) {
        return HIPC_ERR_S_INV_RW_RANGE;
    }

    mdr_p->body_p =
        ((unsigned char *) stru_p) +
        mdr_p->header[HIPC_MSGHDR_BYTE_OFFSET];

    return HIPC_SUCCESS;
}
