#include <string.h>

#include <hipc/atomic.h>
#include <hipc/pbank.h>

/**
 * Initializes a parameter bank.
 * This function copies contents in the first half of buffer to the second half.
 *
 * @param bank_p pointer to a parameter bank
 * @param buf_p a buffer whose size must be \c page_size multiplied by 2
 * @param page_size size of a page
 */
void hipcPbankInit(struct HIPC_pbank *const bank_p,
                   void *const buf_p, const size_t page_size)
{
    bank_p->stat = HIPC_PBANK_STAT_GROUND;
    bank_p->page_size = page_size;
    bank_p->page_ro_p = buf_p;
    bank_p->page_rw_p = (unsigned char *) buf_p + page_size;
    memcpy(bank_p->page_rw_p, buf_p, page_size);
    HIPC_MEM_BARRIER;

    return;
}

/**
 * Gets page size.
 *
 * @param bank_p pointer to a parameter bank
 * @returns page size
 */
size_t hipcPbankGetPageSize(struct HIPC_pbank * const bank_p)
{
    return bank_p->page_size;
}

/**
 * Gets a current read only page.
 *
 * @param bank_p pointer to a parameter bank
 * @returns current read only page
 */
const void *hipcPbankGetReadOnlyPage(struct HIPC_pbank *const bank_p)
{
    return bank_p->page_ro_p;
}

/**
 * Gets a current writable page.
 *
 * @param bank_p pointer to a parameter bank
 * @returns current writable page
 */
void *hipcPbankGetWritablePage(struct HIPC_pbank *const bank_p)
{
    return bank_p->page_rw_p;
}

/**
 * Gets current status of a bank.
 *
 * @param bank_p pointer to a parameter bank
 * @returns current status
 */
enum HIPC_pbank_stat hipcPbankGetStat(struct HIPC_pbank *const bank_p)
{
    return bank_p->stat;
}

/**
 * Requests to switch pages.
 *
 * @param bank_p pointer to a parameter bank
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno hipcPbankSwitchRequest(struct HIPC_pbank *const bank_p)
{
    if (HIPC_PBANK_STAT_GROUND != bank_p->stat) {
        return HIPC_ERR_S_PBANK_STAT_MISMATCH;
    }

    HIPC_MEM_BARRIER;
    bank_p->stat = HIPC_PBANK_STAT_REQUESTED;

    return HIPC_SUCCESS;
}

/**
 * Accepts a switching request and switches pages if switching is requested
 *
 * @param bank_p pointer to a parameter bank
 * @retval 0 Switching is not requested.
 * @retval 1 Pages have been switched.
 */
int hipcPbankSwitchAccept(struct HIPC_pbank *const bank_p)
{
    void *tmp_p;

    if (HIPC_PBANK_STAT_REQUESTED != bank_p->stat) {
        return 0;
    } else {
        /* swap page_ro_p for page_rw_p */
        tmp_p = bank_p->page_ro_p;
        bank_p->page_ro_p = bank_p->page_rw_p;
        bank_p->page_rw_p = tmp_p;

        HIPC_MEM_BARRIER;
        bank_p->stat = HIPC_PBANK_STAT_ACCEPTED;

        return 1;
    }
}

/**
 * Makes both pages have the same latest contents and completes the switching.
 *
 * @param bank_p pointer to a parameter bank
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno hipcPbankSwitchComplete(struct HIPC_pbank *const bank_p)
{
    if (HIPC_PBANK_STAT_ACCEPTED != bank_p->stat) {
        return HIPC_ERR_S_PBANK_STAT_MISMATCH;
    }

    memcpy(bank_p->page_rw_p, bank_p->page_ro_p, bank_p->page_size);
    HIPC_MEM_BARRIER;
    bank_p->stat = HIPC_PBANK_STAT_GROUND;

    return HIPC_SUCCESS;
}

/**
 * Calls hipcPbankSwitchRequest(), waits for acceptance of the request,
 * and calls hipcPbankSwitchComplete().
 *
 * @param bank_p pointer to a parameter bank
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno hipcPbankSwitchReqWaitCmpl(struct HIPC_pbank *const bank_p)
{
    enum HIPC_errno retval;

    retval = hipcPbankSwitchRequest(bank_p);
    if (HIPC_SUCCESS != retval) {
        return retval;
    }

    while (HIPC_PBANK_STAT_REQUESTED == hipcPbankGetStat(bank_p)) {
        ;                       /* nothing to do */
    }

    return hipcPbankSwitchComplete(bank_p);
}
