#include <sys/select.h>
#include <unistd.h>

#include <assert.h>

#include <hipc/posix/ioutils.h>

/**
 * Check a file descriptor is ready to read by using select(2) and 
 * returns the result of select(2).
 *
 * @param fd file descriptor
 * @retval 1 \c fd is at the ready.
 * @retval 0 \c fd is not at the ready.
 * @retval -1 means occurrence of an error.
 */
int hipcPosixReadReady(int fd)
{
    fd_set fds;
    struct timeval timeout;

    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    return select(fd + 1, &fds, NULL, NULL, &timeout);
}

/**
 * Read \c count bytes from \c fd by using read(2).
 * \c count must be less than or equal to \c SSIZE_MAX,
 * otherwise the result is unspecified.
 *
 * @param fd File descriptor
 * @param buf Buffer into which the bytes are read
 * @param count Number of the bytes to be read.
 * @retval >0 The number of the bytes that have been read
 * @retval 0 \c The fd is at the end of the file or \c count is 0.
 * @retval -1 means occurrence of an error, and errno is set by read(2).
 */
ssize_t hipcPosixReadExactly(const int fd, void *buf, const size_t count)
{
    ssize_t n;
    ssize_t i;

    n = count;
    while (0 < n) {
        i = read(fd, buf, n);
        if (0 >= i) {           /* read(2) returns 0 when fd is at the end of the file */
            return i;
        }
        buf = ((unsigned char *) buf) + i;
        assert(n >= i);
        n -= i;
    }
    assert(0 == n);
    return count - n;
}

/**
 * Writes \c count bytes to \c fd by using write(2).
 * \c count must be less than or equal to \c SSIZE_MAX,
 * otherwise the result is unspecified.
 *
 * @param fd File descriptor
 * @param buf Buffer from which bytes are written
 * @param count Number of bytes to be written.
 * @retval >=0 Number of bytes that have been written
 * @retval -1 means occurrence of an error, and errno is set by write(2).
 */
ssize_t hipcPosixWriteExactly(const int fd,
                              void const *buf, const size_t count)
{
    ssize_t n;
    ssize_t i;

    n = count;
    while (0 < n) {
        i = write(fd, buf, n);
        if (0 > i) {
            return i;
        }
        buf = ((unsigned char *) buf) + i;
        assert(n >= i);
        n -= i;
    }
    assert(0 == n);
    return count - n;
}
