#include <string.h>

#include <hipc/usoc.h>
#include <hipc/usocutils.h>

/**
 * Writes \c md_p to \c so_p.
 *
 * @param so_p Pointer to a universal socket
 * @param md_p Pointer to HIPC message data
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno
hipcUsocWriteMsgd(struct HIPC_usoc *const so_p,
                  hipc_msgd const *const md_p)
{
    enum HIPC_errno retval;

    retval = hipcUsocWrite(so_p, hipcMsgdHdrPtr(md_p), HIPC_MSGHDR_SIZE);
    if (HIPC_SUCCESS != retval) {
        return retval;
    }
    retval =
        hipcUsocWrite(so_p, hipcMsgdBdyPtr(md_p), hipcMsgdBdySize(md_p));
    if (HIPC_SUCCESS != retval) {
        return retval;
    }

    return HIPC_SUCCESS;
}

/**
 * Sends system messages of \c cfg_p.
 *
 * @param so_p Pointer to a universal socket
 * @param cfg_p Pointer to a configuration
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
static enum HIPC_errno
hipcUsocSendSysmsg(struct HIPC_usoc *const so_p,
                   const struct HIPC_scfg *const cfg_p)
{
    enum HIPC_errno retval;
    unsigned int i;

    for (i = 0; i < (cfg_p)->n_sysmsg; ++i) {
        retval = hipcUsocWrite(so_p,
                               (cfg_p)->sysmsg_p[i].ptr,
                               (cfg_p)->sysmsg_p[i].size);
        if (HIPC_SUCCESS != retval) {
            return retval;
        }
    }

    return HIPC_SUCCESS;
}

/**
 * Accepts a HIPC session if \c cfg_ary includes a configuration of coming HELLO message.
 * Accepted configuration is set to \c cfg_pp.
 *
 * @param so_p Pointer to a universal socket
 * @param cfg_pp Configuration accepted
 * @param cfg_ary Acceptable configurations
 * @param n_cfg Number of acceptable configurations
 * @param buf buffer for HELLO message
 * @param buf_size size of the buffer
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno
hipcUsocAccept(struct HIPC_usoc *const so_p,
               const struct HIPC_scfg * *const cfg_pp,
               const struct HIPC_scfg *const *const cfg_ary,
               const unsigned int n_cfg,
               unsigned char *const buf, const size_t buf_size)
{
    enum HIPC_errno retval;
    unsigned char body_size;
    unsigned int i;

    *cfg_pp = NULL;

    if (NULL == cfg_ary) {
        return HIPC_ERR_S_INV_NULL;
    }

    /* receive cfgid */
    if (HIPC_MSGHDR_SIZE > buf_size) {
        return HIPC_ERR_S_BUF_SHTG;
    }
    retval = hipcUsocRead(so_p, buf, HIPC_MSGHDR_SIZE);
    if (HIPC_SUCCESS != retval) {
        return retval;
    }
    body_size = hipcMsghdrBdysize(buf);
    if ((HIPC_MSGHDR_SIZE + (size_t) body_size) > buf_size) {
        return HIPC_ERR_S_BUF_SHTG;
    }
    retval = hipcUsocRead(so_p, buf + HIPC_MSGHDR_SIZE, body_size);
    if (HIPC_SUCCESS != retval) {
        return retval;
    }

    /* find cfgid */
    for (i = 0; i < n_cfg; ++i) {
        if (NULL == cfg_ary[i]) {
            return HIPC_ERR_S_INV_NULL;
        }
        if (0 != memcmp(cfg_ary[i]->cfgid_header, buf, HIPC_MSGHDR_SIZE)) {
            continue;
        }
        if (0 ==
            memcmp(cfg_ary[i]->cfgid, buf + HIPC_MSGHDR_SIZE, body_size)) {
            *cfg_pp = cfg_ary[i];
            break;
        }
    }
    if (NULL == *cfg_pp) {
        return HIPC_ERR_S_INV_CFG;
    }

    hipcScfgInit(*cfg_pp);
    retval = hipcUsocSendSysmsg(so_p, *cfg_pp);
    if (HIPC_SUCCESS != retval) {
        return retval;
    }

    return hipcScfgCheckStrusize(*cfg_pp);
}

/**
 * Accepts a HIPC session of \c cfg_p.
 * This function use the stack for a buffer,
 * so the stack size must be large enough.
 *
 * @param so_p Pointer to a universal socket
 * @param cfg_p Acceptable configuration
 * @returns HIPC_SUCCESS on success, or an error on failure.
 */
enum HIPC_errno
hipcUsocAcceptSimple(struct HIPC_usoc *const so_p,
                     const struct HIPC_scfg *const cfg_p)
{
    enum HIPC_errno retval;
    unsigned char body_size;
    unsigned char c;
    unsigned char i;

    if (NULL == cfg_p) {
        return HIPC_ERR_S_INV_NULL;
    }

    for (i = 0; i < HIPC_MSGHDR_SIZE; ++i) {
        retval = hipcUsocRead(so_p, &c, 1);
        if (HIPC_SUCCESS != retval) {
            return retval;
        }
        if (cfg_p->cfgid_header[i] != c) {
            return HIPC_ERR_S_INV_CFG;
        }
    }

    body_size = hipcMsghdrBdysize(cfg_p->cfgid_header);
    for (i = 0; i < body_size; ++i) {
        retval = hipcUsocRead(so_p, &c, 1);
        if (HIPC_SUCCESS != retval) {
            return retval;
        }
        if (cfg_p->cfgid[i] != c) {
            return HIPC_ERR_S_INV_CFG;
        }
    }

    hipcScfgInit(cfg_p);
    retval = hipcUsocSendSysmsg(so_p, cfg_p);
    if (HIPC_SUCCESS != retval) {
        return retval;
    }

    return hipcScfgCheckStrusize(cfg_p);
}
