#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <mock.h>

#include <hipc/usoc.h>
#include <hipc/usocutils.h>

#define S HIPC_SUCCESS
#define R 0
#define W 1

struct HIPC_usoc {
    struct MockIO mio;
};

enum HIPC_errno
hipcUsocRead(struct HIPC_usoc *so_p, void *buf, const size_t size)
{
    return mockIORead(&so_p->mio, buf, size);
}

enum HIPC_errno
hipcUsocWrite(struct HIPC_usoc *so_p, const void *buf, const size_t size)
{
    return mockIOWrite(&so_p->mio, buf, size);
}

static void test_hipcUsocWriteMsgd(void)
{
    struct HIPC_usoc usoc;
    hipc_msgd md;

    {
        struct MockIOOp ops[] = {
            {W, S, 4, {HIPC_MESSAGE_TYPE_QUIT,HIPC_UNSTRU,11,3}},
            {W, S, 3, {0x68,0x69,0x00}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        hipcGenmdQuit(&md, 11, "hi");
        assert(HIPC_SUCCESS == hipcUsocWriteMsgd(&usoc, &md));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {W, S, 4, {HIPC_MESSAGE_TYPE_SUCCESS,HIPC_UNSTRU,0,0}},
            {W, S, 0},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        hipcGenmdSuccess(&md);
        assert(HIPC_SUCCESS == hipcUsocWriteMsgd(&usoc, &md));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {W, S, 4, {HIPC_MESSAGE_TYPE_QUIT,HIPC_UNSTRU,11,3}},
            {W, HIPC_ERR_IO_WRITE, 3, {0x68,0x69,0x00}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        hipcGenmdQuit(&md, 11, "hi");
        assert(HIPC_ERR_IO_WRITE == hipcUsocWriteMsgd(&usoc, &md));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {W, HIPC_ERR_IO_WRITE, 4, {HIPC_MESSAGE_TYPE_QUIT,HIPC_UNSTRU,11,3}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        hipcGenmdQuit(&md, 11, "hi");
        assert(HIPC_ERR_IO_WRITE == hipcUsocWriteMsgd(&usoc, &md));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }

    return;
}

static void test_hipcUsocAccept(void)
{
    struct HIPC_usoc usoc;
    struct MOCK1S1 mstru;
    {
        static unsigned char buf[HIPC_MSGHDR_SIZE +
                                 HIPCUD_cfgmock1_CFGID_SIZE];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO,HIPC_UNSTRU,0,9}},
            {R, S, 9, {0x43,0x49,0x44,0x5f,0x4d,0x4f,0x43,0x4b,0x31}},
            {W, S, 5, {HIPC_MESSAGE_TYPE_SYS,108,0,1,
                       sizeof(mstru)}},
            {W, S, 6, {HIPC_MESSAGE_TYPE_SYS,0,1,2,
                       offsetof(struct MOCK1S1,m1),offsetof(struct MOCK1S1,m2)}},
            {W, S, 6, {HIPC_MESSAGE_TYPE_SYS,0,2,2,
                       sizeof(mstru.m1),sizeof(mstru.m2)}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_SUCCESS==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(&HIPCUD_cfgmock1_SCFG == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] =
            { &HIPCUD_cfgmock2_SCFG, &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO,HIPC_UNSTRU,0,9}},
            {R, S, 9, {0x43,0x49,0x44,0x5f,0x4d,0x4f,0x43,0x4b,0x31}},
            {W, S, 5, {HIPC_MESSAGE_TYPE_SYS,108,0,1,
                       sizeof(mstru)}},
            {W, S, 6, {HIPC_MESSAGE_TYPE_SYS,0,1,2,
                       offsetof(struct MOCK1S1,m1),offsetof(struct MOCK1S1,m2)}},
            {W, S, 6, {HIPC_MESSAGE_TYPE_SYS,0,2,2,
                       sizeof(mstru.m1),sizeof(mstru.m2)}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_SUCCESS==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 2, buf, sizeof(buf)));
        assert(&HIPCUD_cfgmock1_SCFG == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        assert(HIPC_ERR_S_INV_NULL==hipcUsocAccept(&usoc, &cfg_accepted, NULL, 0, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
    }
    {
        static unsigned char buf[1];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        assert(HIPC_ERR_S_BUF_SHTG==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, HIPC_ERR_IO_READ, 4, {HIPC_MESSAGE_TYPE_HELLO,HIPC_UNSTRU,0,9}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_IO_READ==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[4];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO, HIPC_UNSTRU, 0, 9}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_S_BUF_SHTG==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[255];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO, HIPC_UNSTRU, 0, 254}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_S_BUF_SHTG==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO, HIPC_UNSTRU, 0, 9}},
            {R, HIPC_ERR_IO_READ, 9, {0x43,0x49,0x44,0x5f,0x4d,0x4f,0x43,0x4b,0x31}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_IO_READ==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO, HIPC_UNSTRU, 0, 1}},
            {R, S, 1, {0x43}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert( HIPC_ERR_S_INV_CFG==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] =
            { NULL, &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO,HIPC_UNSTRU,0,9}},
            {R, S, 9, {0x43,0x49,0x44,0x5f,0x4d,0x4f,0x43,0x4b,0x31}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_S_INV_NULL==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 2, buf, sizeof(buf)));
        assert(NULL == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        static unsigned char buf[16];
        const struct HIPC_scfg *cfg_accepted;
        const struct HIPC_scfg *cfg_ary[] = { &HIPCUD_cfgmock1_SCFG };
        struct MockIOOp ops[] = {
            {R, S, 4, {HIPC_MESSAGE_TYPE_HELLO,HIPC_UNSTRU,0,9}},
            {R, S, 9, {0x43,0x49,0x44,0x5f,0x4d,0x4f,0x43,0x4b,0x31}},
            {W, HIPC_ERR_IO_WRITE, 5, {HIPC_MESSAGE_TYPE_SYS,108,0,1,
                       sizeof(mstru)}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops)/sizeof(ops[0]));
        assert(HIPC_ERR_IO_WRITE==hipcUsocAccept(&usoc, &cfg_accepted, cfg_ary, 1, buf, sizeof(buf)));
        assert(&HIPCUD_cfgmock1_SCFG == cfg_accepted);
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }

    return;
}

static void test_hipcUsocAcceptSimple(void)
{
    struct HIPC_usoc usoc;
    struct MOCK1S1 mstru;
    {
        struct MockIOOp ops[] = {
            {R, S, 1, {HIPC_MESSAGE_TYPE_HELLO}},
            {R, S, 1, {HIPC_UNSTRU}},
            {R, S, 1, {0}},
            {R, S, 1, {9}},
            {R, S, 1, {0x43}},
            {R, S, 1, {0x49}},
            {R, S, 1, {0x44}},
            {R, S, 1, {0x5f}},
            {R, S, 1, {0x4d}},
            {R, S, 1, {0x4f}},
            {R, S, 1, {0x43}},
            {R, S, 1, {0x4b}},
            {R, S, 1, {0x31}},
            {W, S, 5, {HIPC_MESSAGE_TYPE_SYS,108,0,1,
                       sizeof(mstru)}},
            {W, S, 6, {HIPC_MESSAGE_TYPE_SYS,0,1,2,
                       offsetof(struct MOCK1S1,m1),offsetof(struct MOCK1S1,m2)}},
            {W, S, 6, {HIPC_MESSAGE_TYPE_SYS,0,2,2,
                       sizeof(mstru.m1),sizeof(mstru.m2)}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_SUCCESS ==
               hipcUsocAcceptSimple(&usoc, &HIPCUD_cfgmock1_SCFG));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        mockIOInit(&usoc.mio, NULL, 0);
        assert(HIPC_ERR_S_INV_NULL == hipcUsocAcceptSimple(&usoc, NULL));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {R, HIPC_ERR_IO_READ, 1, {HIPC_MESSAGE_TYPE_HELLO}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_IO_READ ==
               hipcUsocAcceptSimple(&usoc, &HIPCUD_cfgmock1_SCFG));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {R, S, 1, {HIPC_MESSAGE_TYPE_HELLO}},
            {R, S, 1, {HIPC_UNSTRU}},
            {R, S, 1, {0}},
            {R, S, 1, {8}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_S_INV_CFG ==
               hipcUsocAcceptSimple(&usoc, &HIPCUD_cfgmock1_SCFG));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {R, S, 1, {HIPC_MESSAGE_TYPE_HELLO}},
            {R, S, 1, {HIPC_UNSTRU}},
            {R, S, 1, {0}},
            {R, S, 1, {9}},
            {R, HIPC_ERR_IO_READ, 1, {0x43}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_IO_READ ==
               hipcUsocAcceptSimple(&usoc, &HIPCUD_cfgmock1_SCFG));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {R, S, 1, {HIPC_MESSAGE_TYPE_HELLO}},
            {R, S, 1, {HIPC_UNSTRU}},
            {R, S, 1, {0}},
            {R, S, 1, {9}},
            {R, S, 1, {0x43}},
            {R, S, 1, {0x00}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_S_INV_CFG ==
               hipcUsocAcceptSimple(&usoc, &HIPCUD_cfgmock1_SCFG));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }
    {
        struct MockIOOp ops[] = {
            {R, S, 1, {HIPC_MESSAGE_TYPE_HELLO}},
            {R, S, 1, {HIPC_UNSTRU}},
            {R, S, 1, {0}},
            {R, S, 1, {9}},
            {R, S, 1, {0x43}},
            {R, S, 1, {0x49}},
            {R, S, 1, {0x44}},
            {R, S, 1, {0x5f}},
            {R, S, 1, {0x4d}},
            {R, S, 1, {0x4f}},
            {R, S, 1, {0x43}},
            {R, S, 1, {0x4b}},
            {R, S, 1, {0x31}},
            {W, HIPC_ERR_IO_WRITE, 5, {HIPC_MESSAGE_TYPE_SYS,108,0,1,
                                       sizeof(mstru)}},
        };
        mockIOInit(&usoc.mio, ops, sizeof(ops) / sizeof(ops[0]));
        assert(HIPC_ERR_IO_WRITE ==
               hipcUsocAcceptSimple(&usoc, &HIPCUD_cfgmock1_SCFG));
        assert(1 == mockIOCheckCompletion(&usoc.mio));
    }

    return;
}

int main(int argc, char *argv[])
{
    if (2 != argc) {
        fprintf(stderr, "Usage: %s test_id\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    switch (argv[1][0]) {
    case 'a':
        test_hipcUsocWriteMsgd();
        break;
    case 'b':
        test_hipcUsocAccept();
        break;
    case 'c':
        test_hipcUsocAcceptSimple();
        break;
    default:
        fprintf(stderr, "Invalid test_id: %s \n", argv[1]);
        exit(EXIT_FAILURE);
    }

    return 0;
}
